# main.py
#
# Copyright 2023 Letterpress Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")

from gi.repository import Adw, Gdk, Gio, GLib, Gtk

from .file_chooser import FileChooser
from .pasting import Paster
from .window import LetterpressWindow
from .tips_dialog import TipsDialog

from .profile import APP_ID, PROFILE


class LetterpressApplication(Adw.Application):
    """The main application singleton class."""

    def __init__(self):
        super().__init__(
            application_id=APP_ID,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE | Gio.ApplicationFlags.HANDLES_OPEN,
            resource_base_path="/io/gitlab/gregorni/Letterpress",
        )
        self.__create_action("quit", lambda *args: self.quit(), ["<primary>q"])
        self.__create_action("tips", self.__on_tips_action)
        self.__create_action("about", self.__on_about_action)
        self.__create_action(
            "open-file",
            lambda *args: self.get_active_window().on_open_file(),
            ["<primary>o"],
        )
        self.__create_action(
            "zoom-out",
            lambda *args: self.get_active_window().zoom(zoom_out=True),
            ["<primary>minus", "<primary>KP_Subtract"],
        )
        self.__create_action(
            "zoom-in",
            lambda *args: self.get_active_window().zoom(),
            ["<primary>plus", "<primary>KP_Add"],
        )
        self.__create_action(
            "reset-zoom",
            lambda *args: self.get_active_window().reset_zoom(),
            ["<primary>0", "<primary>KP_0"],
        )
        self.__create_action(
            "increase-output-width",
            lambda *args: self.__change_output_width(),
            ["<primary><alt>plus"],
        )
        self.__create_action(
            "decrease-output-width",
            lambda *args: self.__change_output_width(True),
            ["<primary><alt>minus"],
        )
        self.__create_action("paste-image", self.__paste_image, ["<primary>v"])
        self.__create_action(
            "copy-output", self.__copy_output_to_clipboard, ["<primary>c"]
        )
        self.__create_action(
            "save-output",
            lambda *args: FileChooser.save_file(self.get_active_window()),
            ["<primary>s"],
        )
        self.__create_action(
            "open-output", self.__open_output, param=GLib.VariantType("s")
        )
        self.add_main_option_entries([])
        self.file = None
        self.pasted_tmpfile = None
        self.converted_tmpfile = None

    def do_activate(self):
        """Called when the application is activated.

        We raise the application's main window, creating it if
        necessary.
        """
        win = self.get_active_window()
        if win == None:
            win = LetterpressWindow(application=self)
        win.present()

        if PROFILE == "development":
            win.add_css_class("devel")

        if self.file != None:
            win.check_is_image(Gio.File.new_for_path(self.file))

    def do_open(self, files, _n_files, _hint):
        self.file = files[0].get_path()
        self.activate()

    def do_shutdown(self):
        self.converted_tmpfile = None

        if self.pasted_tmpfile:
            try:
                self.pasted_tmpfile.delete()
            except GLib.Error:
                pass

        Gtk.Application.do_shutdown(self)

    def __paste_image(self, *args):
        win = self.get_active_window()
        Paster().paste_image(win, win.check_is_image)

    def __change_output_width(self, down=False):
        win = self.get_active_window()
        if win.filepath != None:
            spin_btn = win.width_spin
            spin_btn.set_value(spin_btn.get_value() + (-100 if down else 100))

    def __copy_output_to_clipboard(self, *args):
        win = self.get_active_window()
        output_text = win.output_label.get_label()
        Gdk.Display.get_default().get_clipboard().set(output_text)
        win.toast_overlay.add_toast(Adw.Toast(title=_("Output copied to clipboard")))

    def __open_output(self, app, data):
        file = open(data.unpack(), "r")
        Gio.DBusProxy.new_sync(
            connection=Gio.bus_get_sync(Gio.BusType.SESSION, None),
            flags=Gio.DBusProxyFlags.NONE,
            info=None,
            name="org.freedesktop.portal.Desktop",
            object_path="/org/freedesktop/portal/desktop",
            interface_name="org.freedesktop.portal.OpenURI",
            cancellable=None,
        ).call_with_unix_fd_list_sync(
            method_name="OpenFile",
            parameters=GLib.Variant(
                "(sha{sv})", ("", 0, {"ask": GLib.Variant("b", True)})
            ),
            flags=Gio.DBusCallFlags.NONE,
            timeout_msec=-1,
            fd_list=Gio.UnixFDList.new_from_array([file.fileno()]),
            cancellable=None,
        )

    def do_command_line(self, command_line):
        args = command_line.get_arguments()
        if len(args) > 1:
            file = command_line.create_file_for_arg(args[1])
            self.file = file.get_path()
        self.activate()
        return 0

    def __on_about_action(self, *args):
        """Callback for the app.about action."""
        about = Adw.AboutDialog.new_from_appdata(
            "/io/gitlab/gregorni/Letterpress/metainfo.xml", "2.1"
        )
        about.set_artists(
            [
                "Brage Fuglseth https://bragefuglseth.dev",
                "kramo https://kramo.hu",
            ]
        )
        # These are Python lists: Add your string to the list (separated by a comma)
        # See the translator comment below for possible formats
        about.set_developers(["Gregor Niehl https://gitlab.gnome.org/gregorni"])
        about.set_copyright(_("Copyright © 2023 Letterpress Contributors"))
        # Translators: Translate this string as your translator credits.
        # Name only:    Gregor Niehl
        # Name + URL:   Gregor Niehl https://gitlab.gnome.org/gregorni/
        # Name + Email: Gregor Niehl <gregorniehl@web.de>
        # Do not remove existing names.
        # Names are separated with newlines.
        about.set_translator_credits(_("translator-credits"))

        about.add_acknowledgement_section(
            _("Code and Design borrowed from"),
            [
                "Upscaler https://gitlab.gnome.org/World/Upscaler",
                "Frog https://github.com/TenderOwl/Frog",
            ],
        )

        # about.add_legal_section(
        #     title="artem",
        #     copyright="Copyright © 2022 artem contributors",
        #     license_type=Gtk.License.MPL_2_0,
        # )

        about.add_legal_section(
             title="jp2a",
             copyright="Copyright 2006-2016 Christian Stigen Larsen and 2020 Christoph Raitzig",
             license_type=Gtk.License.GPL_2_0,
        )

        about.present(self.get_active_window())

    def __on_tips_action(self, *args):
        TipsDialog().present(self.get_active_window())

    def __create_action(self, name, callback, shortcuts=None, param=None):
        """Add an application action.

        Args:
            name: the name of the action
            callback: the function to be called when the action is
              activated
            shortcuts: an optional list of accelerators
            param: an optional list of parameters for the action
        """
        action = Gio.SimpleAction.new(name, param)
        action.connect("activate", callback)
        self.add_action(action)
        if shortcuts:
            self.set_accels_for_action(f"app.{name}", shortcuts)


def main(version):
    """The application's entry point."""
    return LetterpressApplication().run(sys.argv)



# pasting.py
#
# Copyright 2023 Letterpress Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gdk, Gio

from . import texture_to_file


class Paster:
    def paste_image(self, parent_window, callback):
        self.parent_window = parent_window
        self.callback = callback

        clipboard = Gdk.Display.get_default().get_clipboard()
        clipboard.read_value_async(Gio.File, 0, None, self.__on_file_pasted)

    def __on_file_pasted(self, clipboard, result):
        try:
            paste_as_file = clipboard.read_value_finish(result)
            self.callback(paste_as_file)

        except:
            clipboard.read_texture_async(None, self.__on_texture_pasted)

    def __on_texture_pasted(self, clipboard, result):
        try:
            paste_as_texture = clipboard.read_texture_finish(result)
            file = texture_to_file.to_file(paste_as_texture)
            app = self.parent_window.get_application()
            if app.pasted_tmpfile:
                try:
                    app.pasted_tmpfile.delete()
                except GLib.Error:
                    pass
            app.pasted_tmpfile = file
            self.callback(file)

        except:
            toast = Adw.Toast.new(_("No image found in clipboard"))
            self.parent_window.toast_overlay.add_toast(toast)



# zoom_box.py
#
# Copyright 2024 Letterpress Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .zoom_consts import INITIAL_ZOOM, ZOOM_FACTOR, MIN_ZOOM, MAX_ZOOM


@Gtk.Template(resource_path="/io/gitlab/gregorni/Letterpress/gtk/zoom-box.ui")
class ZoomBox(Gtk.Box):
    __gtype_name__ = "ZoomBox"

    zoom_indicator = Gtk.Template.Child()
    decrease_btn = Gtk.Template.Child()
    increase_btn = Gtk.Template.Child()

    def set_zoom(self, zoom_level):
        # map our actual zoom level to a human readable zoom level between 10% and 100%
        display_zoom = (zoom_level - MIN_ZOOM) / (MAX_ZOOM - MIN_ZOOM) * 90 + 10

        self.zoom_indicator.set_label(f"{int(display_zoom)}%")
        self.decrease_btn.set_sensitive(zoom_level > MIN_ZOOM)
        self.increase_btn.set_sensitive(zoom_level < MAX_ZOOM)



# window.py
#
# Copyright 2023 Letterpress Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import subprocess
from tempfile import NamedTemporaryFile

from gi.repository import Adw, Gdk, Gio, Gtk
from PIL import Image, ImageChops, ImageOps

from . import texture_to_file
from .file_chooser import FileChooser
from .zoom_box import ZoomBox
from .zoom_consts import INITIAL_ZOOM, ZOOM_FACTOR, MIN_ZOOM, MAX_ZOOM


@Gtk.Template(resource_path="/io/gitlab/gregorni/Letterpress/gtk/window.ui")
class LetterpressWindow(Adw.ApplicationWindow):
    __gtype_name__ = "LetterpressWindow"

    menu_btn = Gtk.Template.Child()
    drag_revealer = Gtk.Template.Child()
    toast_overlay = Gtk.Template.Child()
    main_stack = Gtk.Template.Child()
    output_scrolled_window = Gtk.Template.Child()
    output_label = Gtk.Template.Child()
    width_spin = Gtk.Template.Child()
    toolbox = Gtk.Template.Child()
    gesture_zoom = Gtk.Template.Child()
    scroll_controller = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.style_manager = Adw.StyleManager.get_default()
        self.style_manager.connect("notify", self.__set_color_scheme)

        target = Gtk.DropTarget(
            formats=Gdk.ContentFormats.new_for_gtype(Gio.File),
            actions=Gdk.DragAction.COPY,
        )

        target.set_gtypes([Gdk.Texture, Gio.File])
        target.connect("drop", self.__on_drop)
        target.connect("enter", self.__on_enter)
        target.connect(
            "leave", lambda *args: self.drag_revealer.set_reveal_child(False)
        )
        self.add_controller(target)

        settings = Gio.Settings(schema_id="io.gitlab.gregorni.Letterpress")
        bind_flags = Gio.SettingsBindFlags.DEFAULT
        settings.bind("window-width", self, "default-width", bind_flags)
        settings.bind("window-height", self, "default-height", bind_flags)
        settings.bind("window-is-maximized", self, "maximized", bind_flags)

        settings.bind(
            "output-width", self.width_spin, "value", Gio.SettingsBindFlags.DEFAULT
        )
        self.width_spin.set_value(settings["output-width"])

        self.width_spin.connect("value-changed", self.__on_spin_value_changed)

        self.gesture_zoom.connect("scale-changed", self.__on_gesture)
        self.scale_delta = 1

        self.scroll_controller.connect("scroll", self.__on_scroll)

        self.previous_stack = "welcome"
        self.filepath = None

        self.zoom_box = ZoomBox()
        self.menu_btn.props.popover.add_child(self.zoom_box, "zoom")
        self.zoom_level = INITIAL_ZOOM
        self.pinch_counter = 0
        self.scrolled_distance = 0

    def on_open_file(self):
        self.main_stack.set_visible_child_name("spinner-page")
        FileChooser.open_file(self, self.previous_stack)

    def check_is_image(self, file):
        filepath = file.get_path()

        def __wrong_image_type():
            print(f"{filepath} is not of a supported image type.")
            # Translators: Do not translate "{basename}"
            toast_text = _("“{basename}” is not of a supported image type.").format(
                basename=file.get_basename()
            )
            self.toast_overlay.add_toast(Adw.Toast.new(toast_text))
            self.main_stack.set_visible_child_name(self.previous_stack)

        try:
            with Image.open(filepath) as img:
                self.main_stack.set_visible_child_name("spinner-page")
                print(f"Input file: {filepath}")

                app = self.get_application()
                app.converted_tmpfile = NamedTemporaryFile(suffix=".png")
                self.filepath = app.converted_tmpfile.name

                shrunken_img = ImageOps.cover(img, (500, 500))
                exif_rotated_img = ImageOps.exif_transpose(shrunken_img)
                exif_rotated_img.save(self.filepath, format="PNG")

                self.__convert_image(self.filepath)
                self.reset_zoom()
        except IOError:
            __wrong_image_type()

    def reset_zoom(self):
        self.zoom_level = INITIAL_ZOOM
        self.__apply_zoom()

    def zoom(self, zoom_out=False):
        if not self.zoom_box.get_sensitive():
            return

        zoom_direction = -ZOOM_FACTOR if zoom_out else ZOOM_FACTOR
        clamp_zoom = lambda level: max(MIN_ZOOM, min(level, MAX_ZOOM))
        self.zoom_level = clamp_zoom(self.zoom_level + zoom_direction)
        self.__apply_zoom()

    def __apply_zoom(self):
        # apply zoom by using the CSS `scale` function
        # see https://developer.mozilla.org/en-US/docs/Web/CSS/transform-function/scale
        css_provider = Gtk.CssProvider.new()
        css_provider.load_from_string(
            f"""label {{
            transform: scale({self.zoom_level});
            }}"""
        )
        self.output_label.get_style_context().add_provider(
            css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )
        self.zoom_box.set_zoom(self.zoom_level)

    def __convert_image(self, filepath):
        self.main_stack.set_visible_child_name("spinner-page")

        arguments = ["jp2a", f"--width={int(self.width_spin.get_value())}", filepath]
        # arguments = ["artem", f"--size={int(self.width_spin.get_value())}", filepath]
        if not self.style_manager.get_dark():
            arguments.append("--invert")

        output = subprocess.Popen(
            arguments,
            stdout=subprocess.PIPE,
            universal_newlines=True,
        ).stdout.readline

        joint_output = "".join(line for line in iter(output, ""))
        self.output_label.set_label(joint_output)

        self.toolbox.set_reveal_child(True)
        self.previous_stack = "view-page"
        self.main_stack.set_visible_child_name(self.previous_stack)

        self.zoom_box.set_sensitive(True)

    def __on_gesture(self, gesture, scale, *args):
        self.pinch_counter += 1
        scale_changed = scale != self.scale_delta
        change_is_big_enough = self.pinch_counter >= 6

        if scale_changed and change_is_big_enough:
            self.zoom(zoom_out=scale < self.scale_delta)
            self.scale_delta = scale
            self.pinch_counter = 0

    def __on_scroll(self, scroll, dx, dy, *args):
        current_modifiers = scroll.get_current_event_state()
        ctrl_is_held = current_modifiers in (
            Gdk.ModifierType.CONTROL_MASK,
            (Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.LOCK_MASK),
        )
        device_is_mouse = (
            scroll.get_current_event_device().get_source() == Gdk.InputSource.MOUSE
        )

        if ctrl_is_held and device_is_mouse:
            self.scrolled_distance += abs(dy)

            if self.scrolled_distance >= 1:
                self.zoom(zoom_out=dy > 0)
                self.scrolled_distance = 0

    def __set_color_scheme(self, *args):
        if self.filepath != None:
            self.__convert_image(self.filepath)

    def __on_spin_value_changed(self, spin_button):
        self.main_stack.set_visible_child_name("spinner-page")
        self.__convert_image(self.filepath)

    def __on_enter(self, *args):
        self.drag_revealer.set_reveal_child(True)
        return Gdk.DragAction.COPY

    def __on_drop(self, widget, drop, *args):
        self.drag_revealer.set_reveal_child(False)
        failed_as_file = False
        try:
            self.check_is_image(drop)
        except:
            failed_as_file = True

        if not failed_as_file:
            return

        try:
            file = texture_to_file.to_file(drop)
            app = self.get_application()
            if app.pasted_tmpfile:
                try:
                    app.pasted_tmpfile.delete()
                except GLib.Error:
                    pass
            app.pasted_tmpfile = file
            self.check_is_image(file)
        except:
            toast = Adw.Toast.new(_("Dropped item is not a valid image"))
            self.toast_overlay.add_toast(toast)



# file_chooser.py
#
# Copyright 2023 Letterpress Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gio, GLib, Gtk

from . import supported_formats


class FileChooser:
    """Open and load file."""

    @staticmethod
    def open_file(parent, last_view_page, *args):
        def __on_response(_dialog, response, _dialog_ref):
            """Run if the user selects a file."""
            if response == Gtk.ResponseType.ACCEPT:
                parent.check_is_image(_dialog.get_file())
            else:
                parent.main_stack.set_visible_child_name(last_view_page)

        dialog = Gtk.FileChooserNative.new(
            title=_("Select a file"), parent=parent, action=Gtk.FileChooserAction.OPEN
        )

        dialog.set_modal(True)
        dialog.connect("response", __on_response, dialog)

        file_filter = Gtk.FileFilter.new()
        file_filter.set_name(_("Supported image files"))
        for file_format in supported_formats.formats:
            file_filter.add_mime_type(f"image/{file_format}")
        dialog.add_filter(file_filter)

        dialog.show()

    @staticmethod
    def save_file(parent, *args):
        def __on_save_file(file):
            print(f"Output file: {file.get_path()}")
            text = parent.output_label.get_label()
            if text != None:
                file.replace_contents_bytes_async(
                    contents=GLib.Bytes.new(text.encode("utf-8")),
                    etag=None,
                    make_backup=False,
                    flags=Gio.FileCreateFlags.NONE,
                    cancellable=None,
                    callback=__save_file_complete,
                )

        def __save_file_complete(file, result):
            info = file.query_info(
                "standard::display-name", Gio.FileQueryInfoFlags.NONE
            )
            display_name = (
                info.get_attribute_string("standard::display-name")
                if info != None
                else file.get_basename()
            )

            toast = Adw.Toast(
                # Translators: Do not translate "{display_name}"
                title=_("Unable to save “{display_name}”").format(
                    display_name=display_name
                )
            )
            if not file.replace_contents_finish(result):
                print(f"Unable to save {display_name}")
            else:
                toast.set_title(
                    # Translators: Do not translate "{display_name}"
                    _("“{display_name}” saved").format(display_name=display_name)
                )
                toast.set_button_label(_("Open"))
                toast.set_action_name("app.open-output")
                toast.set_action_target_value(GLib.Variant("s", file.get_path()))
            parent.toast_overlay.add_toast(toast)

        def __on_response(_dialog, response, _dialog_ref):
            """Run if the user selects a file."""
            if response == Gtk.ResponseType.ACCEPT:
                __on_save_file(_dialog.get_file())

        dialog = Gtk.FileChooserNative.new(
            title=_("Select a file"), parent=parent, action=Gtk.FileChooserAction.SAVE
        )

        dialog.set_modal(True)
        dialog.connect("response", __on_response, dialog)
        dialog.set_current_name("output.txt")
        dialog.show()
